#!/Library/Frameworks/Python.framework/Versions/2.7/bin/python
#
# Last updated 16 Jan 2018
#

#==================================================================================================================================================================
# Defines direction solution utilities used by application programs
#
#   celestialCoords3Station(coincidence, coinc_num, coinc_id, coinc_delta, coinc_timestamp, coinc_event):
#                           Given a list of 3 stations in a coincidence, solve for Celestial Coordinates.
#
#   solveAngles3a(BX,BY,BZ, CX,CY,CZ, delta_b, delta_c) 
#       BX,BY,BZ and CX,CY,CZ define positions in an XYZ coordinate frame. Distances are measured in nano-light-seconds. delta_b and delta_c are time differences.
#
#   haToRa(ha, gps_timestamp)       Given hour ange and GPS time, return the Right Ascension.
#
#   timeDelay( ha, dec, x,y,z)      Work out the time delay for a displacement x,y,z from a base station (distances in light seconds)
#                                   for a shower front moving from direction ha,decself. The equation of the shower is p*x + q*y + r*z = dt.
#==================================================================================================================================================================



import re           # Python standard library     
import math         # Python standard library
import numpy as np  # Python numerical library
import scipy.optimize

import datetime
import utilities

from scipy.optimize           import minimize
from sapphire                 import time_util
from sapphire.transformations import clock
from sapphire.api             import API, Network, Station
from utilities                import degToRad, radToDeg, mtonlgt,combinationList,haDecToECEFUnitVec

debug = 1

def coincCoords(data,coincidence, station_to_num,cindx,s_index):
    """Given a coincidence data structure,
        return a list of the stations (by number)
        a list of the station-ids
        the time delays relative to the reference station.
        their GPS locations
        the station event data. 
    """

    coinc_stat_num  = []               # List of station numbers, defining the coincidence.
    coinc_stat_id   = []               # List of station ids (such as "/hisparc/cluster_birmingham/station_14003").

    coinc_gps       = []               # GPS coordinates of detectors
    coinc_event     = []               # List of event ids in the coincidence.
    
    coinc_lat       = []               # Station latitude
    coinc_lng       = []               # Station longitude

    coinc_ecef_x    = []               # Earth-centred-Earth-fixed X coordinate
    coinc_ecef_y    = []               # Earth-centred-Earth-fixed Y coordinate
    coinc_ecef_z    = []               # Earth-centred-Earth-fixed Z coordinate

    coinc_delta     = []               # List of time offsets (nanoseconds) from first detector listed.
    
    coinc_ecef_dx   = []               # X Displacement from first detector listed in nano-light-seconds
    coinc_ecef_dy   = []               # Y Displacement from first detector listed in nano-light-seconds
    coinc_ecef_dz   = []               # Z Displacement from first detector listed in nano-light-seconds

    coinc_enu_dx    = []               # X Displacement from first detector listed in nano-light-seconds
    coinc_enu_dy    = []               # Y Displacement from first detector listed in nano-light-seconds
    coinc_enu_dz    = []               # Z Displacement from first detector listed in nano-light-seconds

    coinc_inc_angle = []               # Included angle between line from detector 1->2 and detector 1->3 
    
    for s in range(len(cindx)):        # Iterate over stations in this coincidence.
                       
        (seq_id,event_no) = cindx[s]   # The seq_id value indexes into s_index[] to return a station id (such as /hisparc/cluster_birmingham/station_14003)

        station_id     = s_index[seq_id]                               # A string, such as "/hisparc/cluster_birmingham/station_14003"
        station_number = station_to_num[station_id]                    # Now get the station number from the dictionary set up by the stationStatToNumDict() utility function
        event_data     = data.get_node(station_id,'events')[event_no]  # And the event data.

        s_nanosecs = event_data['nanoseconds']
        s_delta    = s_nanosecs - coincidence['nanoseconds']

        station  = Station(station_number,force_stale=True)
        location = station.gps_location()                               # A dictionary with GPS information
        lat      = degToRad*location['latitude']
        lng      = degToRad*location['longitude']

        (ecef_x,ecef_y,ecef_z) = utilities.gps_to_ecef( lat,lng, location['altitude'] )          # Aboslute position of detector A in ECEF coordinate system (meters)

        if s == 0:                                  # The first station is used as a reference location
            ecef_dx = 0.0                           # Displacement in ECEF coordinates in X direction from reference etc.
            ecef_dy = 0.0
            ecef_dz = 0.0
            enu_dx  = 0.0                           # Displacement in ENU coordinates in X direction from reference etc.
            enu_dy  = 0.0
            enu_dz  = 0.0
        else:
            ecef_dx = ecef_x - coinc_ecef_x[0]      # Displacement in ECEF coordinates in X direction from reference
            ecef_dy = ecef_y - coinc_ecef_y[0]      # Displacement in ECEF coordinates in Y direction from reference
            ecef_dz = ecef_z - coinc_ecef_z[0]      # Displacement in ECEF coordinates in Z direction from reference
            
            (enu_dx,enu_dy,enu_dz) = utilities.ecef_to_enu( ecef_x,ecef_y,ecef_z, 
                                                            lat, lng, 
                                                            coinc_ecef_x[0],coinc_ecef_y[0],coinc_ecef_z[0])       # displacement from detector A (meters) in East/North/Up coordinate system - using the utility function.
        
        coinc_stat_num.append( station_number )
        coinc_stat_id.append( station_id )
        coinc_delta.append( s_delta )
        coinc_gps.append( location )
        coinc_event.append( event_data )
        
        coinc_lat.append(lat)                      # GPS latitude
        coinc_lng.append(lng)                      # GPS longitude

        coinc_ecef_x.append(ecef_x)                # X coordinate of detector in ECEF coordinates - meters
        coinc_ecef_y.append(ecef_y)                # Y coordinate of detector in ECEF coordinates - meters
        coinc_ecef_z.append(ecef_z)                # Z coordinate of detector in ECEF coordinates - meters

        coinc_ecef_dx.append(mtonlgt*ecef_dx)      # X Displacement from first detector listed in nano-light-seconds
        coinc_ecef_dy.append(mtonlgt*ecef_dy)      # Y Displacement from first detector listed in nano-light-seconds
        coinc_ecef_dz.append(mtonlgt*ecef_dz)      # Z Displacement from first detector listed in nano-light-seconds

        coinc_enu_dx.append(mtonlgt*enu_dx)        # X Displacement from first detector listed in nano-light-seconds (utility func calc)
        coinc_enu_dy.append(mtonlgt*enu_dy)        # Y Displacement from first detector listed in nano-light-seconds (utility func calc)
        coinc_enu_dz.append(mtonlgt*enu_dz)        # Z Displacement from first detector listed in nano-light-seconds (utility func calc)
        
        # End loop on stations in the coincidence.
        #-------------------------------------------------------------------------------------------------------------------------------------
        
        
    # Included angle - calculated in ECEF coordinate system
    ecef_bac  = radToDeg*utilities.includedAngle(coinc_ecef_dx[1], coinc_ecef_dy[1], coinc_ecef_dz[1],  
                                                 coinc_ecef_dx[2], coinc_ecef_dy[2], coinc_ecef_dz[2])

        
    return (coinc_stat_num, coinc_stat_id, coinc_gps, coinc_event,
                     coinc_lat,coinc_lng,
                     coinc_ecef_x, coinc_ecef_y, coinc_ecef_z,
                     coinc_delta,
                     coinc_ecef_dx,coinc_ecef_dy,coinc_ecef_dz,
                     coinc_enu_dx, coinc_enu_dy, coinc_enu_dz,
                     ecef_bac
                     )


#============================================================================================================
def dt_errors(dt,dx,dy,dz,p,q,r):
    """Given a set of arrival times at a set of detectors with displacements
        from the first detector specified in dx,dy,dz lists,
        and a unit vector p,q,r specifying a shower arrival direction,
        return a list of expected arrival times at each detector
        and a root mean square difference between actual and expected arrival times.
    """
    sumsqdt=0.0
    rdt  = []
    for i in range(len(dt)):
        invdt = p*dx[i]+q*dy[i]+r*dz[i]
        rdt.append(invdt)
        sumsqdt=sumsqdt+(dt[i]-invdt)*(dt[i]-invdt)
        
    rms_dt = math.sqrt(sumsqdt/(len(dt)-1))
    
    return (rdt,rms_dt)


def celestialCoordsMultiPerm(dt, dx,dy,dz , lat,lng):
    """Iterate through all the 3-fold combinations of the dt,dx,dy,dz coordinates
        calculating 3-station solutions for each 3-station combination.
        Return a solutions which is based on the average of all the calculated unit vectors.
    """
    
    plist = combinationList(3,len(dt))
    
    psum = 0.0
    qsum = 0.0
    rsum = 0.0
    nsum = 0
    
    if(debug > 4):
        print " "
        
    for prm in range(len(plist)):
        pick = plist[prm]
        
        dt3 = []
        dx3 = []
        dy3 = []
        dz3 = []
        for c in range(3):            
            dt3.append(dt[pick[c]]-dt[pick[0]])
            dx3.append(dx[pick[c]]-dx[pick[0]])
            dy3.append(dy[pick[c]]-dy[pick[0]])
            dz3.append(dz[pick[c]]-dz[pick[0]])    
            
        (ha,dec,alt,azm,bac,F) = celestialCoords3Station( dt3, dx3,dy3,dz3, lat,lng )  
        
        if(abs(radToDeg*bac) > 20.0 and abs(radToDeg*bac)<160.0):
            (p,q,r) = haDecToECEFUnitVec(ha,dec)
            
            (rdt,rms_dt) = dt_errors(dt3,dx3,dy3,dz3,p,q,r)
            
            psum = psum+p
            qsum = qsum+q
            rsum = rsum+r
            nsum = nsum+1
            
            if(debug > 4):
                print "P=%2d"%(prm),
                print " pick:",pick,
                print "  DTs %3d %3d %3d"%(dt3[0],dt3[1],dt3[2]),
                print " RDTs %3d %3d %3d"%(rdt[0],rdt[1],rdt[2]),
                print " RMSDT %9.3f"%(rms_dt),
                print "  DXs %10.5f %10.5f %10.5f"%(dx3[0],dx3[1],dx3[2]),
                print "DYs %10.5f %10.5f %10.5f"%(dy3[0],dy3[1],dy3[2]),
                print "DZs %10.5f %10.5f %10.5f"%(dz3[0],dz3[1],dz3[2]),
                print "BAC %7.3f"%(radToDeg*bac),
                print "   p,q,r: %10.5f %10.5f %10.5f"%(p,q,r), 
                print "   ha,dec: %10.5f %10.5f"%(radToDeg*ha,radToDeg*dec)
        
    if(nsum>0):
        pave = psum/nsum
        qave = qsum/nsum
        rave = rsum/nsum
        
        (rdt,rms_dt) = dt_errors(dt,dx,dy,dz,pave,qave,rave)

        (ha_ave,dec_ave)   = utilities.ECEFUnitVecToHaDec(pave,qave,rave)
        (alt_ave,azm_ave)  = utilities.haDecToAltAzm(ha_ave,dec_ave,  lat,lng)
    
        dvec_norm= math.sqrt(pave*pave +qave*qave +rave*rave)
        
        if(debug >= 4):
            print "Ave (p,q,r)",pave,qave,rave,
            print " ha,dec",radToDeg*ha_ave,radToDeg*dec_ave,
            print " alt,azm",radToDeg*alt_ave,radToDeg*azm_ave,
            print " RMSDT %9.3f"%(rms_dt),
            print " dvec_norm ",dvec_norm
    else:
        ha = 0.0
        dec= 0.0
        alt= 0.0
        azm= 0.0
        rms_dt = 0.0
        
    return (ha,dec,alt,azm,rms_dt)



def celestialCoordsMultiStation(dt, dx,dy,dz,lat,lng):
    """Given a list of >3 stations in a coincidence, solve for Celestial Coordinates.
       Use a direct minimisation of the least squares subject to the constraint p^2+r^2+q^2 = 1

        Input:
            dt  list of times of arrival with respect to the arrival time at the first detector
            dx  x displacements (nano-light-seconds) of detectors wrt first detector
            dy  y displacements (nano-light-seconds) of detectors wrt first detector
            dz  z displacements (nano-light-seconds) of detectors wrt first detector

        Output: a list containing
            gmst                GMT time stamp for the coincidence. (Seconds since the epoch.)
            ra1,ha1,dec1,       Right ascension, hour angle, and declination of the shower arrival direction (radians)
            alt1,azm1,          Altitude and Azimuth - radians (N.B. azimuth measured from North, coming round first to East.)
            bac,                Included angle BAC betweent the detector arms (radians)
            Flag                Blank if no issues identified during the calculation.
    """

    global debug   

    #----------------------------------------------------------------------------
    # Internal function definition to make use of lexical scoping of dt,dx,dy,dz
    def lqfunc(dvec,sign=1.0):
        """The function to be minimised.
           N.B. p=dvec[0],q=dvec[1],r=dvec[2]
            The function is defined as:
                Sum_i [  ((p*dx(i)+q*dy(i)+r*dz(i))-dt(i))^2 ]
        """  
        sumsq = 0.0
        
        for i in range(len(dt)):
            ta = dvec[0]*dx[i]+dvec[1]*dy[i]+dvec[2]*dz[i]
            sumsq = sumsq + (ta-dt[i])*(ta-dt[i])
            
        return sign*sumsq
        
    #---------------------------------------------------------------------------
    def lqfunc_deriv(dvec,sign=1.0):
        """Derivatives of the function to be minimised, with respect to p,q,r
           N.B. p=dvec[0],q=dvec[1],r=dvec[2]        
            The function is defined as:
                Sum_i [  ((p*dx(i)+q*dy(i)+r*dz(i))-dt(i))^2 ]
            So,
                dFunc/dp = 2*Sum_i[p*dx(i)^2 + q*dx(i)*dy(i) + r*dx(i)*dz(i) - p*dx(i)*dt(i)]
                dFunc/dq = 2*Sum_i[p*dx(i)*dy(i) + q*dx(i)^2 + r*dy(i)*dz(i) - q*dy(i)*dt(i)]
                dFunc/dr = 2*Sum_i[p*dx(i)*dz(i) + q*dy(i)*dz(i) + r*dz(i)^2 - r*dz(i)*dt(i)]        """  
        
        p = dvec[0]
        q = dvec[1]
        r = dvec[2]

        ndx = np.array(dx)  
        ndy = np.array(dy)
        ndz = np.array(dz)
        ndt = np.array(dt)
    
        sumxx = np.sum(np.dot(ndx,ndx))
        sumyy = np.sum(np.dot(ndy,ndy))
        sumzz = np.sum(np.dot(ndz,ndz))
    
        sumxy = np.sum(np.dot(ndx,ndy))
        sumyz = np.sum(np.dot(ndy,ndz))
        sumzx = np.sum(np.dot(ndz,ndx))
    
        sumxt = np.sum(np.dot(ndx,ndt))
        sumyt = np.sum(np.dot(ndy,ndt))
        sumzt = np.sum(np.dot(ndz,ndt))        
        
        dfdp = sign*2.0*(p*sumxx+q*sumxy+r*sumzx-sumxt)
        dfdq = sign*2.0*(p*sumxy+q*sumyy+r*sumyz-sumyt)
        dfdr = sign*2.0*(p*sumzx+q*sumyz+r*sumzz-sumzt) 
        
        return np.array([dfdp,dfdq,dfdr])
        
    #---------------------------------------------------------------------------
    # Constraint p*p + q*1 + r*r-1.0 = 0.0
    cons = ({'type': 'eq',
    'fun' : lambda x: np.array([ x[0]*x[0] + x[1]*x[1] + x[2]*x[2] - 1.0]),
    'jac' : lambda x: np.array([2.0*x[0],2.0*x[1],2.0*x[2]])   })
    #---------------------------------------------------------------------------
           
    res = minimize(lqfunc, [0.0,0.0,1.0], args=(),jac=lqfunc_deriv, constraints=cons, method='SLSQP',options={'disp': False})
    
    p = res.x[0]
    q = res.x[1]
    r = res.x[2]
    
    dvec      = np.array([p,q,r])
    dvec_norm = np.linalg.norm(dvec)       # Normalised direction vector.
        
    (ha,dec) = utilities.ECEFUnitVecToHaDec(p,q,r)    
    (alt,azm) = utilities.haDecToAltAzm(ha,dec,  lat,lng)
    (rdt,rms_dt) = dt_errors(dt,dx,dy,dz,p,q,r)
    
    
    if(debug >= 4):
            
        print "DTs ",dt,
        print "DXs ",dx,
        print "DYs ",dy,
        print "DZs ",dz,
        print "   p, q, r: ",p,q,r, "  RDTs: ",rdt,
        print "   ha,dec: ",radToDeg*ha,radToDeg*dec,
        print " alt,azm",radToDeg*alt,radToDeg*azm,
        print " RMSDT %9.3f"%(rms_dt),
        


    return (ha,dec,alt,azm,rms_dt)
  
   
#============================================================================================================


def celestialCoords3Station( dt, dx,dy,dz, lat,lng ):
    """Given a list of 3 stations in a coincidence, solve for Celestial Coordinates.

        Input:


        Output: a list containing
            gmst                GMT time stamp for the coincidence. (Seconds since the epoch.)
            ra1,ha1,dec1,       Right ascension, hour angle, and declination of the shower arrival direction (radians)
            alt1,azm1,          Altitude and Azimuth - radians (N.B. azimuth measured from North, coming round first to East.)
            bac,                Included angle BAC betweent the detector arms (radians)
            Flag                Blank if no issues identified during the calculation.
    """

    global debug

    BX = 0
    BY = 0
    BZ = 0
    CX = 0
    CY = 0
    CZ = 0
    
    eMessage = ""

    delta_a = 0.0                                                           # BY DEFINITION!!

    delta_b = dt[1]-dt[0]                                                   # Delay of shower arrival time (nano seconds after arrival at detector A).

    BX = dx[1]-dx[0]                                                        # Express X displacement detector in nano-light-seconds.
    BY = dy[1]-dy[0]                                                        # Express Y displacement detector in nano-light-seconds.
    BZ = dz[1]-dz[0]                                                        # Express Z displacement detector in nano-light-seconds.

    dab  = math.sqrt( dx[1]*dx[1] + dy[1]*dy[1] + dz[1]*dz[1]  )            # Distance from A-->B in nano-light-seconds

    delta_c = dt[2]-dt[0]                                                   # Delay of shower arrival time (nano seconds after arrival at detector A).

    CX = dx[2]-dx[0]                                                        # Express X displacement detector in nano-light-seconds.
    CY = dy[2]-dy[0]                                                        # Express Y displacement detector in nano-light-seconds.
    CZ = dz[2]-dz[0]                                                        # Express Z displacement detector in nano-light-seconds.

    dac  = math.sqrt( dx[2]*dx[2] + dy[2]*dy[2] + dz[2]*dz[2]  )            # Distance from A-->C in nano-light-seconds

    bac  = utilities.includedAngle(BX, BY, BZ,  CX, CY, CZ)                 # Included angle - calculated in ECEF coordinate system

    
    if(abs(delta_b) > abs(dab)):                                                # This cannot be a true coincidence, since the delay is larger than the distance between detectors.
        eMessage="abs(delta_b) > abs(dab) "
        return (0.0, 0.0 ,0.0 ,0.0 ,0.0,0.0 ,eMessage)

    elif(abs(delta_c) > abs(dac)):                                              # This cannot be a true coincidence, since the delay is larger than the distance between detectors.
        eMessage="abs(delta_c) > abs(dac) "
        return (0.0, 0.0 ,0.0 ,0.0 ,0.0 ,0.0,eMessage)
            
    else:                                                                       # A valid coincidence.
        sab = math.acos( delta_b/dab )                                          # Angle the shower makes with the A-B detector cluster arm
        sac = math.acos( delta_c/dac )                                          # Angle the shower makes with the A-C detector cluster arm
        
        if(abs(bac)-abs(sab)-abs(sac) > 0.0):                                   # Clearly, if the angle with AB plus the angle with AC is less than BAC something cannot be right!
            eMessage="abs(bac)-abs(sab)-abs(sac) > 0.0 "
            if(debug>4):
                print "Error: abs(bac)-abs(sab)-abs(sac) > 0.0", delta_b,dab,delta_c,dac, radToDeg*bac, radToDeg*sab, radToDeg*sac, radToDeg*abs(bac),radToDeg*abs(sab),radToDeg*abs(sac)
            return (0.0, 0.0 ,0.0 ,0.0 ,0.0,0.0 ,eMessage)
        else:
            
            #======================================================================================================================================
            # Direction solver
            # ----------------
            # Solve for direction in ECEF coordinate system - not that there are always two possible outcomes, 
            # since for any solution on one side of the BAC plane, the mirror image in the plane makes the same angles and must also be a solution.
            ((ha1,dec1),(ha2,dec2),F3) = solveAngles3( BX,BY,BZ, CX,CY,CZ, delta_b, delta_c )
            #======================================================================================================================================
            
            
            if(F3 == "*"):
                eMessage= "Error from solveAngles3"
                
            if(F3 == "*" and debug >=2):
                debug_s = debug
                debug = 4
                ((ha1,dec1),(ha2,dec2),F) = solveAngles3( BX,BY,BZ, CX,CY,CZ, delta_b, delta_c )
                print "bac: %8.3f  bacn: %8.3f sab: %8.3f sac: %8.3f bac-sab-sac: %8.3f"%(radToDeg*bac,radToDeg*bacn,radToDeg*sab,radToDeg*sac, radToDeg*(abs(bac)-abs(sab)-abs(sac)))
                debug = debug_s
                
            # We also need Altitude/Azimuth to distinguish solutions from the Sky from those coming up from the ground.
            (alt1,azm1)  = utilities.haDecToAltAzm(ha1,dec1,  lat,lng)          # Convert solution 1 to Altitude/Azimuth
            (alt2,azm2)  = utilities.haDecToAltAzm(ha2,dec2,  lat,lng)          # Convert solution 2 to Altitude/Azimuth
            
            # Reverse solution
            dtbR1 = timeDelay(ha1,dec1, BX,BY,BZ)                               # Calculated A-B time delay implied by solution 1.
            dtcR1 = timeDelay(ha1,dec1, CX,CY,CZ)                               # Calculated A-C time delay implied by solution 1.
            
            dtbR2 = timeDelay(ha2,dec2, BX,BY,BZ)                               # Calculated A-B time delay implied by solution 2.
            dtcR2 = timeDelay(ha2,dec2, CX,CY,CZ)                               # Calculated A-C time delay implied by solution 2.
            
            # Set flags if there are discrepancies.
            rms_dt1 = math.sqrt(0.5*((dtbR1-delta_b)*(dtbR1-delta_b) + (dtcR1-delta_c)*(dtcR1-delta_c)))
            rms_dt2 = math.sqrt(0.5*((dtbR2-delta_b)*(dtbR2-delta_b) + (dtcR2-delta_c)*(dtcR2-delta_c)))
            
            if(rms_dt1 > 1.E-3):
                eMessage = eMessage+" rms_dt1 > 1.E-3;"
            if(rms_dt2 > 1.E-3):
                eMessage = eMessage+" rms_dt2 > 1.E-3;"
            
            
            if( debug >= 2 ):
                    # Check the the two solutions make the same angle with the detector plane - on opposite sides.
                    (px,py,pz)    = utilities.vectorProduct(dbx,dby,dbz, dcx,dcy,dcz)             # Vector perpendicular to the plane BAC

                    (ux1,uy1,uz1) = utilities.haDecToECEFUnitVec(ha1,dec1)                        # Unit vector of solution 1
                    (ux2,uy2,uz2) = utilities.haDecToECEFUnitVec(ha2,dec2)                        # Unit vector of solution 2

                    a1 = utilities.includedAngle(ux1,uy1,uz1, px, py, pz)                         # Angle between perpendicular and the plane for solution 1.
                    a2 = utilities.includedAngle(ux2,uy2,uz2,-px,-py,-pz)                         # Angle between perpendicular and the plane for solution 2.                
                    print "dtb,dtbR1,dtbR2: %7.2f %7.2f %7.2f  "%(delta_b,dtbR1,dtbR2),
                    print "dtc,dtcR1,dtcR2: %7.2f %7.2f %7.2f  "%(delta_c,dtcR1,dtcR2),            
                    print "%1s%1s   incAngle BAC: %7.2f  "%(F1,F2,radToDeg*bac) ,
                    print "ha1,dec1: %8.2f %7.2f %7.2f  "%(radToDeg*ha1, radToDeg*dec1) ,
                    print "ha2,dec2: %8.2f %7.2f %7.2f  "%(rradToDeg*ha2, radToDeg*dec2),
                    print "alt1/azm1: %7.2f %7.2f  "%(radToDeg*alt1,radToDeg*azm1) ,
                    print "alt2/azm2: %7.2f %7.2f  "%(radToDeg*alt2,radToDeg*azm2) ,
                    print "a1 %7.2f  a2 %7.2f"%(radToDeg*a1,radToDeg*a2)
                    
            if( alt1 >= alt2):
                return (ha1,dec1,alt1,azm1,rms_dt1,bac,eMessage)
            else:
                return (ha1,dec2,alt2,azm2,rms_dt2,bac,eMessage)



def solveAngles3(bx,by,bz, cx,cy,cz, dtb,dtc):
    """
        Used by celestialCoords3Station to do the trigonometry.
    
        Input:
            BX,BY,BZ    displacement of station B relative to station A (nano-light-seconds)
            CX,CY,CZ    displacement of station C relative to station A (nano-light-seconds)
            delta_b     time delay in shower arrival time at station B  (nanoseconds).
            delta_c     time delay in shower arrival time at station C  (nanoseconds).

        Output: a nested list with two hour-angle declination solutions (angles in radians)
            ((ha1,dec1),(ha2,dec2),flag)
        Only one solution will point at the sky, but it may be necessary to look at the solution in Alt/Azm or ENU coordinates to distinguish which.
    """

    if( debug >= 4 ):
        print "BX ,BY ,BZ :",bx,by,bz," CX ,CY ,CZ :",cx,cy,cz, " delta_b: ",dtb," delta_c: ",dtc

    if(debug >= 4):
        print "R (bx*cy-cx*by)=",(bx*cy-cx*by),
        print "P (by*cz-cy*bz)=",(by*cz-cy*bz),
        print "Q (bx*cz-cx*bz)=",(bx*cz-cx*bz)
        
    if( abs(bx*cy-cx*by) > abs(by*cz-cy*bz) ):
        if( abs(bx*cy-cx*by) > abs(bx*cz-cx*bz) ):
            flag = 'R'
        else:
            if( abs(bx*cz-cx*bz) > abs(by*cz-cy*bz) ):
                flag = 'Q'
            else:
                flag = 'P'
    else:
        if( abs(bx*cz-cx*bz) > abs(by*cz-cy*bz) ):
            flag = 'Q'
        else:
            flag = 'P'       

    if ( flag == 'R' ):
        # Solving for R
        upr = -(bz*cy  - cz*by)/(bx*cy-cx*by)
        vpr = (dtb*cy - dtc*by)/(bx*cy-cx*by)
    
        uqr = -( bz*cx -  cz*bx)/(cx*by-bx*cy)
        vqr =  (dtb*cx - dtc*bx)/(cx*by-bx*cy)

        ( r1,r2,ef ) = solveQuad( upr,vpr, uqr, vqr )
        p1 = upr*r1 + vpr
        q1 = uqr*r1 + vqr
        p2 = upr*r2 + vpr
        q2 = uqr*r2 + vqr

    if ( flag == 'Q' ):
        # Solving for Q
        upq = -(by*cz  - cy*bz)/(bx*cz-cx*bz)
        vpq = (dtb*cz - dtc*bz)/(bx*cz-cx*bz)
    
        urq = -( by*cx -  cy*bx)/(bz*cx-cz*bx)
        vrq =  (dtb*cx - dtc*bx)/(bz*cx-cz*bx)

        ( q1,q2,ef ) = solveQuad( upq,vpq, urq,vrq )
        p1 = upq*q1 + vpq
        r1 = urq*q1 + vrq
        p2 = upq*q2 + vpq
        r2 = urq*q2 + vrq

    if( flag == 'P' ):
        # Solving for P
        uqp = -( bx*cz -  cx*bz)/(by*cz-cy*bz)
        vqp =  (dtb*cz - dtc*bz)/(by*cz-cy*bz)

        urp = -( bx*cy - cx*by)/(bz*cy-cz*by)
        vrp =  (dtb*cy -dtc*by)/(bz*cy-cz*by)


        ( p1,p2,ef ) = solveQuad( uqp,vqp, urp,vrp)

        q1 = uqp*p1 + vqp
        r1 = urp*p1 + vrp
        q2 = uqp*p2 + vqp
        r2 = urp*p2 + vrp    


    if(debug >= 4):
        print "Flag:",flag," ef:",ef," P1,Q1,R1: ",p1,q1,r1," P2,Q2,R2: ", p2,q2,r2

    if( ef == " "):        

        (ha1,dec1) = utilities.ECEFUnitVecToHaDec(p1,q1,r1)
        (ha2,dec2) = utilities.ECEFUnitVecToHaDec(p2,q2,r2)

        if(debug >= 4):
            print " ha1,dec1:",radToDeg*ha1,radToDeg*dec1," ha2,dec2:",radToDeg*ha2,radToDeg*dec2
        
        return ((ha1,dec1),(ha2,dec2),flag)
    else:
        return ((0,0),(0,0),"error")


def solveQuad( u1,v1, u2,v2 ):
    """ Solve a quadratic equation for the coordinate solution."""

    A = (u1*u1 + u2*u2 + 1.0)
    B = 2.0*(u1*v1 + u2*v2)
    C = v1*v1+v2*v2 - 1.0
    D = B*B-4.0*A*C

    if( D < 0.0 ):
        if(D > -1.E-10):
            D = 0.0
        if( D < 0.0):
            if(debug >= 2):
                print "A,B,C ",A,B,C," Error! D=",D
            return (0,0,'error')

    s1 = (-B + math.sqrt(D))/(2.0*A)
    s2 = (-B - math.sqrt(D))/(2.0*A)

    return (s1,s2," ")



def haToRa(ha, gps_timestamp):
    """
        Given hour angle of a celestial object, or direction, and GPS time, return the Right Ascension.
    
        Note that Hour Angle is the angle between two planes
        1)  The plane containing the Earth's axis and the Zenith direction
        2)  The plane containing the Earth's axis and the direction to a celestial object.
        Hour Angle increases to the West. (When measured in time units it is therefore the time after the object has crossed the meridian.)
    """

    gmst = clock.gps_to_lst(gps_timestamp, 0.0)  # GMST in decimal hours

    ra1 = gmst*math.pi*2.0/24.0 - ha             # Right Ascension (in radians). 

    ra = (ra1+2.0*math.pi)%(2.0*math.pi)         # Ensure positive and in the (0..2*pi) range

    return (ra,gmst)


def timeDelay( ha, dec, x,y,z):
    """
        Work out the time delay for a displacement x,y,z from a reference detector for a shower front 
        moving FROM direction ha,dec (hour angle and declination, measured in radians).
    
        It is assumed that the x,y,z coordinates are ECEF.

        The equation of the shower front motion is dt = -(p*x + q*y + r*z),
            where (p,q,r) defines a unit vector in the chosen coordinate system pointing TO the direction ha,dec.
            and (x,y,z) define a point in the ECEF coordinate system (relative to the reference detector), measured in light travel time.
        (Hence, if the unit of distance measurement is nano-light-seconds, dt is returned in nano-light-seconds.)
    
       """

    (p,q,r) = utilities.haDecToECEFUnitVec(ha,dec)    # The unit vector returned hear points FROM the source

    dt = +(x*p + y*q + r*z)                           # so we have a positive sign here

    return dt    

   
def testSolveCoords3Station():
    
    dt = [0,        239,        241]  
    dx = [0.00000,  366.23169,  209.87003]
    dy = [0.00000,  624.26286, 1055.98473]
    dz = [0.00000, -277.87914, -212.65053]   
    
    lat = 52.3572524*degToRad
    lng = 4.948385*degToRad
    
    (ha,dec,alt,azm,bac,flag) = celestialCoords3Station(dt,dx,dy,dz,lat,lng)
    (p,q,r) = haDecToECEFUnitVec(ha,dec)
    
    invDt = []
    invDt2= []
    for i in range(3):
        invDt.append(timeDelay(ha,dec,dx[i],dy[i],dz[i]))
        invDt2.append(p*dx[i]+q*dy[i]+r*dz[i])
        
    print dx,dy,dz,dt,invDt,invDt2,p,q,r,radToDeg*ha,radToDeg*dec


if __name__ == "__main__":
    import sys,re,math
   
    testSolveCoords3Station()

